﻿
using System;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics.PackedVector;

namespace Silo {

	public class Pen {
		public Vector4 Color;
		public float Radius;
		public Vector2 Position;
		public bool Pressed;
	}

	public class SiloState {
		readonly Effect effect;
		readonly EffectParameter penColor;
		readonly EffectParameter penRadius;
		readonly EffectParameter penPosition;
		readonly EffectParameter resolution;
		readonly EffectParameter massTexture;
		readonly EffectParameter pressureTexture;
		readonly EffectParameter flowTexture;

		public SiloState(Effect effect) {
			this.effect = effect;
			penRadius = effect.Parameters["penRadius"];
			penPosition = effect.Parameters["penPosition"];
			penColor = effect.Parameters["penColor"];
			resolution = effect.Parameters["resolution"];
			massTexture = effect.Parameters["MassTexture"];
			pressureTexture = effect.Parameters["PressureTexture"];
			flowTexture = effect.Parameters["FlowTexture"];
		}
		public float PenRadius { get { return penRadius.GetValueSingle(); } set { penRadius.SetValue(value); } }
		public Vector2 PenPosition { get { return penPosition.GetValueVector2(); } set { penPosition.SetValue(value); } }
		public Vector4 PenColor { get { return penColor.GetValueVector4(); } set { penColor.SetValue(value); } }
		public Vector2 Resolution { get { return resolution.GetValueVector2(); } set { resolution.SetValue(value); } }
		public Texture2D MassTexture { get { return massTexture.GetValueTexture2D(); } set { massTexture.SetValue(value); } }
		public Texture2D PressureTexture { get { return pressureTexture.GetValueTexture2D(); } set { pressureTexture.SetValue(value); } }
		public Texture2D FlowTexture { get { return flowTexture.GetValueTexture2D(); } set { flowTexture.SetValue(value); } }
	}
	public class SiloField : DrawableGameComponent {
		SpriteBatch sprites;
		Effect effect;
		Buffer2D mass = new Buffer2D();
		Buffer2D massBack = new Buffer2D();
		readonly Buffer2D pressure = new Buffer2D();
		readonly Buffer2D flow = new Buffer2D();
		VertexBuffer vb;
		VertexDeclaration declaration;
		IndexBuffer ib;
		readonly Random random = new Random();
		HalfVector4[] data;
		readonly Pen pen = new Pen();
		EffectPass EllipsePass;
		EffectPass SoftEllipsePass;
		EffectPass PressurePass;
		EffectPass FlowPass;
		EffectPass MovePass;
		EffectPass ColorPass;
		SiloState state;
		int width = 800;
		int height = 600;
		public event EventHandler Resized;

		public SiloField(Game game, int width, int height)
			: base(game) {
			this.width = width;
			this.height = height;
		}

		public SiloState State { get { return state; } }
		public Pen Pen { get { return pen; } }
		public int Width { get { return width; } }
		public int Height { get { return height; } }

		protected override void LoadContent() {
			sprites = new SpriteBatch(GraphicsDevice);
			effect = Game.Content.Load<Effect>("effect/Silo");
			EffectTechnique technique = effect.CurrentTechnique;
			EllipsePass = technique.Passes["Ellipse"];
			SoftEllipsePass = technique.Passes["SoftEllipse"];
			PressurePass = technique.Passes["Pressure"];
			FlowPass = technique.Passes["Flow"];
			MovePass = technique.Passes["Move"];
			ColorPass = technique.Passes["Color"];

			state = new SiloState(effect);

			declaration = new VertexDeclaration(GraphicsDevice, new VertexElement[] {
					new VertexElement(0, 0, VertexElementFormat.HalfVector4, VertexElementMethod.Default, VertexElementUsage.Position, 0),
				});
			vb = new VertexBuffer(GraphicsDevice, typeof(HalfVector4), 3, BufferUsage.None);
			ib = new IndexBuffer(GraphicsDevice, typeof(ushort), 3, BufferUsage.None);
			vb.SetData(new HalfVector4[] {
#if XBOX // Xbox版バグ対処（1.0 Refreshまでの対策です。2.0での動作は未検証です）
					new HalfVector4(1, 0, 0, 0),
					new HalfVector4(1, 0, 0, 2),
					new HalfVector4(1, 0, 2, 0),
#else
					new HalfVector4(0, 0, 0, 1),
					new HalfVector4(2, 0, 0, 1),
					new HalfVector4(0, 2, 0, 1),
#endif
				});
			ib.SetData(new ushort[] { 0, 1, 2 });

			Initialize(width, height);

			base.LoadContent();
		}
		public void Initialize(int width, int height) {
			this.width = width;
			this.height = height;
			state.Resolution = new Vector2(width, height);
			mass.Dispose();
			massBack.Dispose();
			pressure.Dispose();
			flow.Dispose();
			mass.Surface = new RenderTarget2D(GraphicsDevice, width, height, 1, SurfaceFormat.HalfVector4, RenderTargetUsage.PlatformContents);
			massBack.Surface = new RenderTarget2D(GraphicsDevice, width, height, 1, SurfaceFormat.HalfVector4, RenderTargetUsage.PlatformContents);
			pressure.Surface = new RenderTarget2D(GraphicsDevice, width, height, 1, SurfaceFormat.HalfVector4, RenderTargetUsage.PlatformContents);
			flow.Surface = new RenderTarget2D(GraphicsDevice, width, height, 1, SurfaceFormat.HalfVector4, RenderTargetUsage.PlatformContents);
			state.PressureTexture = pressure.Texture;
			state.FlowTexture = flow.Texture;
			data = new HalfVector4[width * height];
			Reset();
			if (Resized != null)
				Resized(this, EventArgs.Empty);
		}
		public void Reset() {
			for (int i = 0; i < width * height; i++) {
#if XBOX // Xbox版バグ対処（1.0 Refreshまでの対策です。2.0での動作は未検証です）
				data[i] = new HalfVector4(
					1,
					(float)random.Next(1),
					(float)Math.Max(0, random.Next(6) - 4),
					(float)Math.Max(0, random.Next(6) - 4));
#else
				data[i] = new HalfVector4(
					(float)Math.Max(0, random.Next(6) - 4),
					(float)Math.Max(0, random.Next(6) - 4),
					(float)random.Next(1),
					1);
#endif
			}
			mass.Texture.SetData(data);
		}

		public override void Update(GameTime gameTime) {
			GraphicsDevice.RenderState.DepthBufferEnable = false;

			state.MassTexture = mass.Texture;
			state.PressureTexture = pressure.Texture;
			state.FlowTexture = flow.Texture;

			state.PenColor = pen.Color;
			state.PenPosition = pen.Position;
			state.PenRadius = pen.Radius;

			effect.Begin();

			GraphicsDevice.VertexDeclaration = declaration;
			GraphicsDevice.Vertices[0].SetSource(vb, 0, declaration.GetVertexStrideSize(0));
			GraphicsDevice.Indices = ib;

			if (pen.Pressed) {
				Fill(EllipsePass, massBack);
				Swap(ref mass, ref massBack);
				state.MassTexture = mass.Texture;
			}

			Fill(PressurePass, pressure);

			Fill(FlowPass, flow);
			if (true) {
				Fill(MovePass, massBack);
				Swap(ref mass, ref massBack);
			}
			effect.End();

			GraphicsDevice.SetRenderTarget(0, null);

			base.Update(gameTime);
		}
		public override void Draw(GameTime gameTime) {
			GraphicsDevice.RenderState.DepthBufferEnable = false;

			state.MassTexture = mass.Texture;
			//state.MassTexture = flow.Texture;
			effect.Begin();

			GraphicsDevice.VertexDeclaration = declaration;
			GraphicsDevice.Vertices[0].SetSource(vb, 0, declaration.GetVertexStrideSize(0));
			GraphicsDevice.Indices = ib;

			Fill(ColorPass, (RenderTarget2D)null);
			effect.End();

			base.Draw(gameTime);
		}
		public void Fill(EffectPass pass, RenderTarget2D target) {
			GraphicsDevice.SetRenderTarget(0, target);
			pass.Begin();
			GraphicsDevice.DrawIndexedPrimitives(PrimitiveType.TriangleStrip, 0, 0, 3, 0, 1);
			pass.End();
			GraphicsDevice.SetRenderTarget(0, null);
		}
		public void Fill(EffectPass pass, Buffer2D target) {
			Fill(pass, target.Surface);
			target.Resolve();
		}
		public static void Swap(ref Buffer2D left, ref Buffer2D right) {
			Buffer2D temp = left;
			left = right;
			right = temp;
		}
	}

}
